/*
 *  Copyright (C) 2013-2014 Katharina Breininger
 *
 *   This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.WindowManager;
import ij.gui.GenericDialog;
import ij.gui.Line;
import ij.gui.Plot;
import ij.gui.PlotWindow;
import ij.gui.PointRoi;
import ij.gui.PolygonRoi;
import ij.gui.Roi;
import ij.measure.Calibration;
import ij.measure.ResultsTable;
import ij.plugin.filter.PlugInFilter;
import ij.plugin.frame.RoiManager;
import ij.process.ByteProcessor;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;
import ij.process.ShortProcessor;
import ij.text.TextWindow;

import java.awt.Color;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Arrays;
import java.util.LinkedList;

import edu.stanford.rsl.jpop.FunctionOptimizer;
import edu.stanford.rsl.jpop.FunctionOptimizer.OptimizationMode;
import edu.stanford.rsl.jpop.FunctionOptimizer.TerminationCode;

/**
 * This Plugin can be used to track the nasal cartilages and the nasal septum during breathing in an MR image sequence. <br>
 * <b>Input</b>: image stack picturing a cross section through the human head over time during breathing. <br>
 * <b>User input</b>: line selection that denotes where the movement should be tracked. <br>
 * <b>Output</b>: resliced image that holds an image line for each point in time. In this image the tracked nasal walls are shown with an ImageJ ROI. Furthermore a table with the position relative to the line selection and distances between the nasal walls. <br>
 * <br>
 * Please note that this plugin can only handle image stacks where all images have the same dimensions.
 * It can only handle line rois at the moment.
 * RGB images are transformed to 8-bit grayscale images (based on luminance). (usual way does not work because curip.getLine(..) uses the method getPixelValue(..), which returns the luminance of a RGB-pixel)
 * @author Katharina Breininger
 *
 */
public class Alae_Tracker implements PlugInFilter, ActionListener {
	/** original input image */
	ImagePlus img;
	/** scaled input image */
	ImagePlus scaledImg;
	/** resliced image that is used for tracking */
	ImagePlus reslicedImg;
	
	/** Interpolation method used for upscaling */
	private int interpolationMethod = ImageProcessor.BILINEAR;
	/** ROI type used for representing the tracking result */
	private ROITypes roiType = ROITypes.POLYGON_ROI;

	private double scaleFactor;
	private boolean scalePicture;
	
	/** Title of the resliced image */
	private String title;

	private boolean interpolateImprobableResults;
	private boolean orderEstimatedResults = true;
	private boolean showPlots = false;
	
	RoiManager roiManager;
	ResultsTable resultsTable;
	
	// ROIs for the left wall, septum, right wall
	Roi left, middle, right;
	
	// Strings for ResultsTable Columns
	protected static final String POSITION_LEFT_SIDE = "Position linker Nasenflgel";
	protected static final String POSITION_RIGHT_SIDE = "Position rechter Nasenflgel";
	protected static final String POSITION_NASAL_SEPTUM = "Position Nasenscheidewand";
	protected static final String DISTANCE_LEFT = "Abstand links";
	protected static final String DISTANCE_RIGHT = "Abstand rechts";
	protected static final String TIME = "Zeitpunkt";
	protected static final String MenuUpdate = "Update";
	
		
	/* ******** SECTION: ImageJ / Interface methods ******** */
	@Override
	public int setup(String arg, ImagePlus imp) {
		if(imp == null) {
			IJ.error("No image with selection open");
			return DONE;
		}
		img = imp;
		return DOES_8G + DOES_16 + DOES_32 + DOES_RGB + ROI_REQUIRED + NO_CHANGES;
	}

	@Override
	public void run(ImageProcessor ip) {
		Roi regionOfInterest = img.getRoi();
		if(regionOfInterest == null) {
			IJ.error("No region of interest defined.");
			return;
		}
		if(regionOfInterest.getType() == Roi.LINE) { 
			if(!showDialog())
				return;
			if(scalePicture) {
				scaledImg = Scaler_.createRescaledStack(img, scaleFactor, interpolationMethod, title, true);
				scaledImg.setRoi(Scaler_.scaleRoi(regionOfInterest, scaleFactor, true, true), false);
				reslicedImg = getReslicedPictureFromLineRoi();
			}
			else {
				scaledImg = img;
				reslicedImg = getReslicedPictureFromLineRoi();
			}
		}
		else{
			IJ.error("Plugin only handles line rois.");
			return;
		}
		reslicedImg.show();

		int[][] anatomicalMaxima = getNasenflugelEstimates(true);
		// get ROIs and put them in the image
		Roi[] anatomicalRoi = getAnatomicalRoi(anatomicalMaxima);
		if(roiManager == null) {
			if(RoiManager.getInstance() == null)
				roiManager = new RoiManager();
			else 
				roiManager = RoiManager.getInstance();
		}
		// set ROIs, names
		left = anatomicalRoi[0];
		left.setName("ROI left wall");
		roiManager.addRoi(left);
		middle = anatomicalRoi[1];
		middle.setName("ROI middle wall");
		roiManager.addRoi(middle);
		right = anatomicalRoi[2];
		right.setName("ROI right wall");
		roiManager.addRoi(right);
		
		// put results in a table
		if(resultsTable == null)
			resultsTable = new ResultsTable();
		putResultsInTable(anatomicalMaxima);
		resultsTable.showRowNumbers(false);
		resultsTable.show("Measurements");
		
		// add Update-option to results table
		TextWindow tw = (TextWindow)WindowManager.getFrame("Measurements");
		MenuBar resultBar = tw.getMenuBar();
		Menu m = new Menu(MenuUpdate);
		m.add(new MenuItem(MenuUpdate));
		m.addActionListener(this);
		resultBar.add(m);
		tw.setMenuBar(resultBar);
	}
	
	
	private void putResultsInTable(int[][] anatomicalMaxima) {
		//get scale
		Calibration reslicedCalibr = reslicedImg.getCalibration();
		for(int i = 0; i < anatomicalMaxima[0].length; i++) {
			resultsTable.incrementCounter();
			resultsTable.addValue(TIME, i);
			resultsTable.addValue(POSITION_LEFT_SIDE, anatomicalMaxima[0][i]);
			resultsTable.addValue(POSITION_NASAL_SEPTUM, anatomicalMaxima[1][i]);
			resultsTable.addValue(POSITION_RIGHT_SIDE, anatomicalMaxima[2][i]);
			resultsTable.addValue(DISTANCE_LEFT, (anatomicalMaxima[1][i]-anatomicalMaxima[0][i])*reslicedCalibr.pixelWidth);
			resultsTable.addValue(DISTANCE_RIGHT, (anatomicalMaxima[2][i]-anatomicalMaxima[1][i])*reslicedCalibr.pixelWidth);
		}
	}

	/**
	 * This method opens a dialog where the parameters for this plugin can be adapted by the user.
	 * @return Returns false, if the dialog has been canceled, true otherwise.
	 */
	private boolean showDialog() {
		scalePicture = true;
		scaleFactor = 10.0;
		title = WindowManager.getUniqueName(img.getTitle());
		String[] methods = ImageProcessor.getInterpolationMethods();

		GenericDialog gd = new GenericDialog("Plugin options");
		gd.addCheckbox("Use a scaled picture for extimation process.", scalePicture);
		gd.addStringField("Scale factor", "" + scaleFactor);
		gd.addChoice("Interpolation method for scaling", methods, methods[interpolationMethod]);
		gd.addCheckbox("Interpolate failed estimation results", true);
		String[] roiTypes = ROITypes.getStringArray();
		gd.addChoice("ROI type for displaying the estimates", roiTypes, roiTypes[ROITypes.POLYGON_ROI.ordinal()]);
		gd.addCheckbox("Show Example Plots", false);
		gd.showDialog();

		if (gd.wasCanceled())
			return false;
		scalePicture = gd.getNextBoolean();
		String scaleStr = gd.getNextString(); // out here to not introduce funny behavior if further text fields are added
		interpolationMethod = gd.getNextChoiceIndex();
		interpolateImprobableResults = gd.getNextBoolean();
		roiType = ROITypes.getByNumber(gd.getNextChoiceIndex());
		showPlots = gd.getNextBoolean();
		if(scalePicture) {
			if(!scaleStr.equals(""+scaleFactor)) {
				try {
					scaleFactor = Double.parseDouble(scaleStr);
				} catch(NumberFormatException e) {
					IJ.showMessage("Invalid input for scale factor - no scaling is used.");
					scaleFactor = 1.0;
				}
			}
			if (scaleFactor < 1.0) {
				IJ.showMessage("Scale factor for this Plugin has to be >= 1.0 - no scaling is used.");
				scaleFactor = 1.0;				
			}
			if(scaleFactor == 1.0) {
				scalePicture = false;
			}
		}
		
		return true;
	}
		
	/* ************ SECTION: Image transformation methods ************* */
	/**
	 * This method transforms the image/pixel coordinates that describe the optimization results/anatomical landmarks
	 * into an array of ROI-Elements. <br>
	 * Both sides and the middle wall of the nose make up a ROI each, that contain a ROI-point for each image in the image stack. <br>
	 * So the first ROI describes the movement of the left nasal wall, the second the movement of the middle nasal wall,
	 * the third the movement of the right nasal wall.
	 * @param anatomicalMaxima The array that contains the estimates. It has the following form: For each image i there are three values for the three anatomical landmarks.
	 * They have the following position in the array: anatomicalMaxima[0][i], anatomicalMaxima[1][i], anatomicalMaxima[2][i].
	 * @return The array with the resulting ROIs described above. The first ROI in the array consists of all values in anatomialMaxima[0] 
	 * - that means contains all estimates for the right nasal wall, for the second and the third element corresponds the middle and the left wall respectively.
	 */
	private Roi[] getAnatomicalRoi(int[][] anatomicalMaxima) {
		int[] xvaluesRoi1 = Arrays.copyOf(anatomicalMaxima[0], anatomicalMaxima[0].length);
		int[] xvaluesRoi2 = Arrays.copyOf(anatomicalMaxima[1], anatomicalMaxima[1].length);
		int[] xvaluesRoi3 = Arrays.copyOf(anatomicalMaxima[2], anatomicalMaxima[2].length);
		int[] yvalues = fillIntegerArray(reslicedImg.getHeight(), 0, 1);
		Roi roi1 = null, roi2 = null, roi3 = null;
		switch(roiType) {
		case POINT_ROI:
			roi1 = new PointRoi(xvaluesRoi1, yvalues, yvalues.length);
			roi2 = new PointRoi(xvaluesRoi2, yvalues, yvalues.length);
			roi3 = new PointRoi(xvaluesRoi3, yvalues, yvalues.length);
			break;
		case FREEHAND_ROI:
			roi1 = new PolygonRoi(xvaluesRoi1, yvalues, yvalues.length, Roi.FREELINE);
			roi2 = new PolygonRoi(xvaluesRoi2, yvalues, yvalues.length, Roi.FREELINE);
			roi3 = new PolygonRoi(xvaluesRoi3, yvalues, yvalues.length, Roi.FREELINE);
			break;
		case POLYGON_ROI:
			roi1 = new PolygonRoi(xvaluesRoi1, yvalues, yvalues.length, Roi.POLYLINE);
			roi2 = new PolygonRoi(xvaluesRoi2, yvalues, yvalues.length, Roi.POLYLINE);
			roi3 = new PolygonRoi(xvaluesRoi3, yvalues, yvalues.length, Roi.POLYLINE);
			break;
		}
		return new Roi[] {roi1, roi2, roi3};
	}
	
	/**
	 * This method is responsible for the actual reslicing: 
	 * Depending on the image type the pixels long the line-roi are read out throughout the stack and 
	 * put together line by line to form the resliced image.
	 * Please note that reslicing a RGB-image results in an 8-bit grayscale image (this is due to the image processing methods in class {@link ImageProcessor}).
	 * 
	 * @param roi The region of interest that specifies the line along which the image is to be resliced.
	 * @return The resliced image.
	 */
	private ImagePlus getReslicedPictureFromLineRoi() {
		Line roi = (Line) scaledImg.getRoi();
		ImageStack stack = scaledImg.getStack();
		int stacksize = stack.getSize();
		ImageProcessor curip = stack.getProcessor(1);
		int linelength =  curip.getLine(roi.x1, roi.y1, roi.x2, roi.y2).length; // because roi.getPixels().length does not give the correct number of pixels for very short ROIs
		
		// Program focused on speed but not on code readability - maybe restructure, also unnecessary array allocation.
		byte[] resultArrayLine8G = new byte[linelength*stacksize];
		short[] resultArrayLine16G = new short[linelength*stacksize];
		float[] resultArrayLine32G = new float[linelength*stacksize];

		// Composition of pixels along the roi-line for each picture in the stack
		for(int i = 0; i < stacksize; ++i) {
			int arrayOffset = i*linelength;
			curip = stack.getProcessor(i+1);
			double[] curline = curip.getLine(roi.x1, roi.y1, roi.x2, roi.y2);
			if(curline.length != linelength)
				IJ.error("Plugin implementation error: Length of current line segment does not comply with linelength before!" + "\nlinelength: "+ linelength + ", curlinelength: "+ curline.length);
			for(int j = 0; j < linelength; ++j) {
				switch(img.getType()){
				case ImagePlus.GRAY8:
				case ImagePlus.COLOR_RGB:
					resultArrayLine8G[j+arrayOffset] = (byte) curline[j];
					break;
				case ImagePlus.GRAY16:
					resultArrayLine16G[j+arrayOffset] = (short) curline[j];
					break;
				case ImagePlus.GRAY32:
					resultArrayLine32G[j+arrayOffset] = (float) curline[j];
					break;
				default:
					IJ.error("Unhandled image type in switch-case statement.");
					throw new IllegalArgumentException("Unhandled image type in switch-case statement.");
				}
			}
		}
		
		// transformation of array into a new ImagePlus
		ImagePlus result;
		switch(img.getType()){
		case ImagePlus.GRAY8:
		case ImagePlus.COLOR_RGB:
			result =  new ImagePlus("Resliced", new ByteProcessor(linelength, stacksize, resultArrayLine8G));
			break;
		case ImagePlus.GRAY16:
			result =  new ImagePlus("Resliced", new ShortProcessor(linelength, stacksize, resultArrayLine16G, null));
			break;
		case ImagePlus.GRAY32:
			result = new ImagePlus("Resliced", new FloatProcessor(linelength, stacksize, resultArrayLine32G));
			break;
		default:
			IJ.error("Unhandled image type in switch-case statement.");
			throw new IllegalArgumentException("Unhandled image type in switch-case statement.");
		}

		// Get the new image calibration:
		// If the RoI is slanted, the pixel width of the resliced line is different to the (scaled) picture.
		// So to get the correct pixel width, the org. pixel width (of the whole scaled image) is multiplied 
		// by the ratio of a resliced line parallel to x-axis and the length of the actual resliced line (= array length).
		Calibration imageCalibration = scaledImg.getCalibration();
		imageCalibration.setXUnit(img.getCalibration().getXUnit());
		
		double xdistance = Math.abs(roi.x2-roi.x1)+1; //length of a resliced line parallel to x-axis
		imageCalibration.pixelWidth = imageCalibration.pixelWidth*(xdistance/linelength);
		result.setCalibration(imageCalibration);
	
		return result;
	}

	/* ************ SECTION: Optimization function methods ************* */
	/**
	 * This method handles the optimization and result improvement process based on the extracted image intensities.
	 * A sum of three Gaussian is used to model the distribution of the intensities and 
	 * the JPOP framework is used to optimize the parameters of this model function.
	 * Optionally an interpolation and ordering of the results is performed.
	 * 
	 * @param statusMessage Indicates whether or not a status message shall be printed to the status bar of ImageJ.
	 * @return The resulting estimates for the maxima of the triple Gaussian (that is used modeling the intensities) 
	 * that are assumed to correspond to the anatomical landmarks. The array has the following form: For each image i there are three values for the three anatomical landmarks.
	 * They have the following position in the array: anatomicalMaxima[0][i], anatomicalMaxima[1][i], anatomicalMaxima[2][i].
	 */
	private int[][] getNasenflugelEstimates(boolean statusMessage) {
		int[][] result = new int[3][reslicedImg.getHeight()];
		LinkedList<Pair<GaussianOptimizationFunction, double[]>> optFunctions = constructOptimizationFunctions();
		int n = reslicedImg.getHeight();
		for(int i = 0; i < reslicedImg.getHeight(); i++) {
			if(statusMessage)
				IJ.showStatus("Get estimation result: " + i + "/" + n);
			double[] parameters = doOptimization(optFunctions.get(i).firstElement, optFunctions.get(i).secondElement);
			if(parameters == null) {
				result[0][i] = result[1][i] = result[2][i] = -1;
			}
			else {
				result[0][i] = (int)Math.round(parameters[0]);
				result[1][i] = (int)Math.round(parameters[2]);
				result[2][i] = (int)Math.round(parameters[4]);
			}
			if(i == 0 && showPlots) {
				double[] xValues = fillCont(reslicedImg.getWidth(), 0);
				double[] intensityValues = reslicedImg.getProcessor().getLine(0, i, reslicedImg.getWidth(), i);
				double[] yValues = getModelFunctionValues(parameters, xValues);
				Plot plotTogether = createPlot(xValues, yValues, intensityValues);
				Plot plotIntensities = createPlot(xValues, intensityValues);
				Plot plotModelFunction = createPlot(xValues, yValues);
				plotTogether.show();
				plotIntensities.show();
				plotModelFunction.show();
			}
			if(statusMessage)
				IJ.showProgress(i, n);
		}
		if(interpolateImprobableResults) {
			IJ.showStatus("Interpolating improbable results.");
			IJ.showProgress(0.0);
			interpolateResultList(result);
		}
		if(orderEstimatedResults) {
			orderResults(result);
		}
		if(statusMessage)
			IJ.showProgress(1.0);
		return result;
	}
	
	/**
	 * This method tries to improve the optimization result by interpolating the maxima, if the optimization process has failed. This is denoted by an array entry of < 0.
	 * @param optimizationResult
	 */
	private void interpolateResultList(int[][] optimizationResult) {
		for(int x = 0; x < optimizationResult.length; x++) {
			for(int i = 0; i < optimizationResult[x].length;i++) {
				if(optimizationResult[x][i] < 0) {
					// get values for interpolation
					int interpolationBasis1 = 0, interpolationBasis2 = 0;
					int j = 0;
					while(optimizationResult[x][i+j] < 0) {
						++j;
						if(i+j >= optimizationResult[x].length) { //if you reach the end of the array - break and use only one value for extrapolation
							break;
						}
					}
					if(i+j >= optimizationResult[x].length && i==0) {
						IJ.error("Selected ROI is to short for interpolation or no valid interpolation results found.");
						return;
					}
					else if(i+j >= optimizationResult[x].length) { // reached end of array
						interpolationBasis1 = interpolationBasis2 = optimizationResult[x][i-1];
					}
					else if(i == 0) { //special case: result[k][i] is the first element of array - use only one value for extrapolation
						interpolationBasis1 = interpolationBasis2 = optimizationResult[x][i+j];
					}
					else { // normal case
						interpolationBasis1 = optimizationResult[x][i-1];
						interpolationBasis2 = optimizationResult[x][i+j];
					}
					// do the actual interpolation
					int start = i-1;
					int end = i+j;
					double distance = j+1;
					for(int curpos = i; curpos < i+j; curpos++) {
						optimizationResult[x][curpos] = (int) Math.round(interpolationBasis1 * (distance - (curpos - start))/(distance) + interpolationBasis2 * (distance-(end - curpos))/(distance));
					}
				}
			}
		}
		
	}

	/**
	 * Constructs the optimization problems from the resliced image.
	 * @return A List with all optimization problems. 
	 * The entries of the list are pairs consisting of the optimization function and the respective initial estimates for the estimation process.
	 */
	private LinkedList<Pair<GaussianOptimizationFunction, double[]>> constructOptimizationFunctions() {
		LinkedList<Pair<GaussianOptimizationFunction, double[]>> resultList = new LinkedList<Pair<GaussianOptimizationFunction, double[]>>();
		ImageProcessor processor = reslicedImg.getProcessor();

		double[] xValues = fillCont(reslicedImg.getWidth(), 0);

		for(int i = 0; i < reslicedImg.getHeight(); i++) {
			double[] intensityValues = processor.getLine(0, i, reslicedImg.getWidth(), i);
			GaussianOptimizationFunction optFunction = new ScalableGaussianOptimizationFunction(xValues, intensityValues);
			double[] initialValues = getInitialValues(intensityValues);
			resultList.add(new Pair<GaussianOptimizationFunction, double[]>(optFunction, initialValues));
		}
		return resultList;
	}

	/**
	 * 
	 * @param intensityValues
	 * @return Returns an array with initial values for the optimization process.
	 */
	private double[] getInitialValues(double[] intensityValues) {
		double sixth = intensityValues.length/6;
		double scaleFactor = intensityValues.length*4; //initial value ursprnglich 1000
		double initialValues[] = {sixth, sixth, 3*sixth, sixth, 5*sixth, sixth, scaleFactor, scaleFactor, scaleFactor};

		return initialValues;
	}

	/**
	 * Does an optimization for the given function using the initial estimates.
	 * @param optimizationProblem A Pair of an optimization function and the initial estimates for the parameters.
	 * @return Returns the estimated parameters from the optimization process.
	 */
	private double[] doOptimization(GaussianOptimizationFunction optimizationFunction,  double[] initialValues) {
		if(initialValues.length != 9) {
			throw new IllegalArgumentException("The number of intial values is incorrect");
		}
		FunctionOptimizer fo = new FunctionOptimizer(9);
		fo.setInitialX(initialValues);
		fo.setOptimizationMode(OptimizationMode.Gradient);
		double[] results = fo.optimizeFunction(optimizationFunction);

		if(interpolateImprobableResults) {
			TerminationCode termCode = fo.getTerminationCode();
			IJ.log(termCode.toString());
			
			if(termCode == TerminationCode.OptimalSolution || termCode == TerminationCode.GradientSmall || termCode==TerminationCode.StepsizeSmall || termCode==TerminationCode.LowestPointFound)
				return results;
			else {
				IJ.log("Interpolate estimization result.");
				return null;
			}
		}

		return results;
	}

	/**
	 * Helping method to improve the optimization result: 
	 * Given calculated maxima that are assumed to correspond to anatomical landmarks
	 * it is checked if these maxima are in the right order, 
	 * which means that (result[0][i] < result[1][i] && result[1][i] < result[2][i]) is fulfilled (for every i). <br>
	 * If this condition is not fulfilled, the values are reordered (a permuted order is also a valid optimization result).
	 * @param optimizationResult The result of the optimization process, that are assumed to correspond to anatomical landmarks
	 * @return The reordered optimizationResult.
	 */
	private void orderResults(int[][] optimizationResult) {
		for(int i = 0; i < optimizationResult[0].length; i++) {
			if(! (optimizationResult[0][i] < optimizationResult[1][i] && optimizationResult[1][i] < optimizationResult[2][i])) { // values are in the wrong order
				// just bubble sort
				for(int x = 2; x > 0; x--) {
					for(int y = 0; y < x; y++) {
						if(optimizationResult[y][i] > optimizationResult[y+1][i]) {
							int temp = optimizationResult[y][i];
							optimizationResult[y][i] = optimizationResult[y+1][i];
							optimizationResult[y+1][i] = temp;
						}
					}
				}
			}
		}
	}

	private double[] getModelFunctionValues(double[] parameters, double[] xValues) {
		double[] results = new double[xValues.length]; 
		GaussianOptimizationFunction functionDummy = new ScalableGaussianOptimizationFunction();
		for(int i = 0; i < xValues.length; ++i) {
			results[i] = functionDummy.evaluateBasisFunction(xValues[i], parameters);
		}
		return results;
	}
	
	/* ************ SECTION: Methods concerning output ************ */
	/**
	 * Constructs a plot for the given x- and y-values. Please note that the x-value array and the y-values array should have the same length.
	 * @param xValues The x-values for the plot.
	 * @param yValues The y-values for the plot.
	 * @return
	 */
	private Plot createPlot(double[] xValues, double[] yValues, double[] intensityValues) {
		if(xValues.length != yValues.length)
			throw new IllegalArgumentException("The number of intial values is incorrect");
		Plot plot = new Plot("ReclicedPlot", "position", "gray value", xValues, yValues);
		plot.setColor(Color.red);
		plot.addPoints(xValues, intensityValues, PlotWindow.LINE);
		plot.setColor(Color.blue);
		return plot;
	}

	/**
	 * Constructs a plot for the given x- and y-values. Please note that the x-value array and the y-values array should have the same length.
	 * @param xValues The x-values for the plot.
	 * @param yValues The y-values for the plot.
	 * @return
	 */
	private Plot createPlot(double[] x, double[] fx) {
		if(x.length != fx.length)
			throw new IllegalArgumentException("The number of intial values is incorrect");
		Plot plot = new Plot("ReclicedPlot", "position", "gray value", x, fx);
		plot.setColor(Color.red);
		return plot;
	}

	
	
	/* ********* SECTION: Additional helping methods and classes ************ */
	/**
	 * Returns a double array filled with continuous integer values.
	 * Example: fullCont(4, 0) returns the following array {0.0, 1.0, 2.0, 3.0}
	 * 
	 * @param length the length of the respective array.
	 * @param begin the number from which on the numbers are counted
	 * @return Returns an array filled with continuous integer values, null if a length < 0 is given as a parameter.
	 */
	protected static double[] fillCont(int length, int begin) {
		if(length < 0)
			return null;
		double[] result = new double[length];
		for(int i = 0; i < length; ++i) {
			result[i] = begin+i;
		}
		return result;
	}
	
	/**
	 * Returns a integer array filled with integer values, that differ consecutively by stepsize. <br>
	 * Examples: fullCont(4, 0, 1) returns the following array {0, 1, 2, 3} <br>
	 * Example: fullCont(4, 0, -1) returns the following array {0, -1, -2, -3}
	 * 
	 * @param length The length of the resulting array.
	 * @param begin The first number in the array.
	 * @param stepsize The difference of two consecutive numbers.
	 * 
	 */	
	protected static int[] fillIntegerArray(int length, int begin, int stepsize) {
		if(length < 0)
			return null;
		int[] result = new int[length];
		for(int i = 0; i < length; ++i) {
			result[i] = begin+i*stepsize;
		}
		return result;
	}
	
	/**
	 * Provides generic 2-element tuple.
	 *
	 * @param <U> Class of the first element.
	 * @param <V> Class of the second element.
	 */
	protected class Pair<U, V> {
		U firstElement;
		V secondElement;
		
		Pair(U value1, V value2) {
			firstElement = value1;
			secondElement = value2;
		}
	}
	
	@Override
	public void actionPerformed(ActionEvent evt) {
		String cmd = evt.getActionCommand();
		if(cmd.equals(MenuUpdate)) {
			updateAnatomicalMeasurements();
		}
		else
			throw new IllegalArgumentException("Action command " + cmd + " is not supported by this ActionListener");
	}
	
	/**
	 * Updates the anatomical measurements that are derived from the optimization process and the 
	 * points the point ROIs are made up of to fit the current/changed ROI in the image
	 */
	private void updateAnatomicalMeasurements() {
		// get ROIs
		Roi[] currentRois = roiManager.getRoisAsArray();
		Roi left1 = null;
		Roi middle1 = null;
		Roi right1 = null;
		for(int i = 0; i < currentRois.length; i++) {
			if(left.getName().equals(currentRois[i].getName()))
				left1 = currentRois[i];
			if(right.getName().equals(currentRois[i].getName()))
				right1 = currentRois[i];
			if(middle.getName().equals(currentRois[i].getName()))
				middle1 = currentRois[i];
		}
		resultsTable.reset();
		int[] xpoints0 = left1.getPolygon().xpoints;
		int[] xpoints1 = middle1.getPolygon().xpoints;
		int[] xpoints2 = right1.getPolygon().xpoints;
		putResultsInTable(new int[][]{xpoints0, xpoints1, xpoints2});
		resultsTable.show("Measurements");

	}
}

/** Enum to handle possible types for ROI that is used to represent tracking result */
enum ROITypes {
	POINT_ROI("Point ROI"), FREEHAND_ROI("Freehand ROI"), POLYGON_ROI("Polygon ROI");
	final String name;
	private ROITypes(final String name) {
		this.name = name;
	}
	public static String[] getStringArray() {
		ROITypes[] vals = ROITypes.values();
		String[] strings = new String[vals.length];
		for(int i = 0; i < vals.length; i++) {
			strings[i] = vals[i].toString();
		}
		return strings;
	}
	public String toString() {
		return name;
	}
	public static ROITypes getByNumber(int i) {
		ROITypes[] vals = ROITypes.values();
		return vals[i];
	}
}