/*
 *  Copyright (C) 2013-2014 Katharina Breininger
 *
 *   This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.gui.Roi;
import ij.measure.ResultsTable;
import ij.plugin.PlugIn;
import ij.plugin.frame.RoiManager;

public class Measure_ROIs implements PlugIn {

	ImagePlus img;
	
	RoiManager roiManager;
	ResultsTable resultsTable;
	
	Roi left, middle, right;
	
	boolean debug;
	Roi[] rois = new Roi[3];
	double pixelWidth = 1.0;
	
	// Strings for ResultsTable Columns
	protected static final String POSITION_LEFT_SIDE = "Position linker Nasenflgel";
	protected static final String POSITION_RIGHT_SIDE = "Position rechter Nasenflgel";
	protected static final String POSITION_NASAL_SEPTUM = "Position Nasenscheidewand";
	protected static final String DISTANCE_LEFT = "Abstand links";
	protected static final String DISTANCE_RIGHT = "Abstand rechts";
	protected static final String TIME = "Zeitpunkt";
	protected static final String MenuUpdate = "Update";
	
		
	/* ******** SECTION: ImageJ / Interface methods ******** */
	@Override
	public void run(String arg) {
		if(!showDialog())
			return;
		// put results in a table
		if(resultsTable == null)
			resultsTable = new ResultsTable();
		measureRois();
		resultsTable.show("Measurements");
	}
	
	
	private void putResultsInTable(int[][] anatomicalMaxima, double pixelWidth) {
		//get scale
		for(int i = 0; i < anatomicalMaxima[0].length; i++) {
			resultsTable.incrementCounter();
			resultsTable.addValue(TIME, i);
			resultsTable.addValue(POSITION_LEFT_SIDE, anatomicalMaxima[0][i]);
			resultsTable.addValue(POSITION_NASAL_SEPTUM, anatomicalMaxima[1][i]);
			resultsTable.addValue(POSITION_RIGHT_SIDE, anatomicalMaxima[2][i]);
			resultsTable.addValue(DISTANCE_LEFT, (anatomicalMaxima[1][i]-anatomicalMaxima[0][i])*pixelWidth);
			resultsTable.addValue(DISTANCE_RIGHT, (anatomicalMaxima[2][i]-anatomicalMaxima[1][i])*pixelWidth);
		}
	}

	
	/**
	 * Updates the anatomical measurements that are derived from the optimization process and the 
	 * points the point ROIs are made up of to fit the current/changed ROI in the image
	 */
	private void measureRois() {
		resultsTable.reset();
		int[] xpoints0 = rois[0].getPolygon().xpoints;
		int[] xpoints1 = rois[1].getPolygon().xpoints;
		int[] xpoints2 = rois[2].getPolygon().xpoints;
		putResultsInTable(new int[][]{xpoints0, xpoints1, xpoints2}, pixelWidth);
		resultsTable.show("Measurements");
	}

	public boolean showDialog() {
		GenericDialog gd = new GenericDialog("Plugin options");
		gd.addMessage("Put in field three ROI IDs for measurement, separate elements with \",\"");
		gd.addStringField("ROIs ", "");
		gd.addStringField("Pixelwidth ", "");
		gd.showDialog();
		if(gd.wasCanceled())
			return false;
		String roiString = gd.getNextString();
		String pixelW = gd.getNextString();
		if(pixelW != null && !pixelW.equals("")) {
			pixelWidth = Double.valueOf(pixelW);
		}
		roiString = roiString.replaceAll("\\s", "");
		String[] ids = roiString.split(",");
		if(debug) {
			for(String id: ids) {
				System.out.println(id);
			}
		}
		if(ids.length != 3) {
			IJ.error("This plugin needs 3 rois - plugin aborted");
			return false;
		}
		
		int[] roiIDs = new int[3];
		for(int i = 0; i < 3; i++) {
			try {
				roiIDs[i] = Integer.parseInt(ids[i]);
			}
			catch(NumberFormatException e) {
				IJ.error("Not able to parse "+ ids[i] + ".");
				return false;
			}
		}
		RoiManager roiManager = RoiManager.getInstance() == null ? new RoiManager() : RoiManager.getInstance() ;
		Roi[] currentRois = roiManager.getRoisAsArray();
		for(int i = 0; i < 3; i++) {
			if(currentRois[roiIDs[i]].getType() != Roi.POLYLINE) {
				IJ.error("ROI type must be POLYLINE! - plugin aborted");
				return false;
			}
			rois[i] = currentRois[roiIDs[i]];
		}
		return true;
	}
}