/*
 *  Copyright (C) 2013-2014 Katharina Breininger
 *
 *   This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.Arrays;

/**
 * This class provides an optimizable function for the JPOP framework ({@link http://www5.cs.fau.de/research/software/java-parallel-optimization-package/} or 
 * {@link http://peaks.informatik.uni-erlangen.de/jpop-api/overview-summary.html}.
 * <p>
 * The goal is to model a given set of values with a sum of three scalable Gaussian functions and optimize the parameters of these Gaussian functions via
 * the sum of squared differences.
 * The following parameters are optimized: Mean (\mu_i), standard deviation (\sigma_i) and scaling factor (\alpha_i) for each Gaussian function.
 * <p>
 * <b>Formula: Sum of three Gaussians (in latex):</b><br>
 * g_s(x) = \sum_{k=1}^{3} \alpha_k \mathcal{N}(x; \mu_k, \sigma_k)  <br>
 * <b>Formula: Optimization function: </b> <br>
 * h_s(\theta_s) = \sum_{i = 1}^{N} (f(x_i)-g_s(x_i))^2 --> minimum <br>
 * with \theta_s = (\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3, \alpha_1, \alpha_2, \alpha_3)
 * <p>
 * Except not noted otherwise in the method description, these parameters must have the same order as shown in the line above and will have the same order in the array that contains the final result of the optimization process.
 * <p>
 * Note and TODO:
 * The parameter indices and their position in the resp. arrays might differ because Java arrays start with 0, but in the notation the indices 1 - 3 are used.
 * The descriptions might not be completely consistent due to this reason.
 */
public class ScalableGaussianOptimizationFunction implements GaussianOptimizationFunction {
	double[] yvalues;
	double[] xvalues;
	
	/**
	 * Constructs a new Optimization function. 
	 * If at least one of the value-arrays is null, 
	 * the result is a "dummy function" - 
	 * so you can call evaluateBasisFunction(..) with given parameters to get the 
	 * function value at the corresponding position.
	 * 
	 * @param xvalues (x_i): Array with the positions where you have corresponding measurements (in yvalues)
	 *  - both arrays must have the same length. 
	 * @param yvalues (f(x_i)): y-values/f(x) values corresponding to the positions in xvalues. 
	 */
	public ScalableGaussianOptimizationFunction(double[] xvalues, double[] yvalues) {
		if(yvalues.length != xvalues.length) {
			throw new IllegalArgumentException("Length of x-value array and y-value array is different");
		}
		this.yvalues = yvalues;
		this.xvalues = xvalues;
	}

	
	public ScalableGaussianOptimizationFunction() {
	}

	
	@Override
	public void setNumberOfProcessingBlocks(int number) {
		throw new UnsupportedOperationException("Setting the number of processing blocks is not supported by this function.");
	}

	@Override
	public int getNumberOfProcessingBlocks() {
		return 1;
	}
	/**
	 * Evaluates the optimization function h_s (see class description) for the parameters (= at the position) params for the Gaussian functions.
	 * 
	 * @param params = {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3, \alpha_1, \alpha_2, \alpha_3}
	 */
	public double evaluate(double[] params, int block) {
		if(params.length != 9) {
			throw new IllegalArgumentException("Wrong number of parameters for evaluation!");
		}
		double[] params_1 = Arrays.copyOf(params, 6);
		double[] scalefactors = {params[6], params[7], params[8]};
		return evaluate(params_1, scalefactors, block);
	}
	
	protected double evaluate(double[] params, double[] scalefactors, int block) {
		if(params.length != 6 || scalefactors.length != 3) {
			throw new IllegalArgumentException("Wrong number of parameters for evaluation!");
		}
		double leastSquareResult = 0;
		for(int i = 0; i < yvalues.length; ++i) {
			leastSquareResult += (yvalues[i]-evaluateTripleGaussian(xvalues[i], params, scalefactors))*(yvalues[i]-evaluateTripleGaussian(xvalues[i], params, scalefactors));
		}
		return leastSquareResult;
	}
	
	@Override
	public double[] gradient(double[] params, int block) {
		if(params.length != 9) {
			throw new IllegalArgumentException("Wrong number of paramters for gradient calculation!");
		}
		double[] params_1 = Arrays.copyOf(params, 6);
		double[] scalefactors = {params[6], params[7], params[8]};

		double[] result = new double[9];
		for(int i = 0; i < 3; ++i) {
			result[i*2] = calculateGradientRespToMean(params_1, scalefactors, i);
			result[i*2+1] = calculateGradientRespToStandardDev(params_1, scalefactors, i);
			result[6+i] = calculateGradientRespToAlpha(params_1, scalefactors, i);
		}
		return result;
	}

	/**
	 * Evaluate the model function g_s (see class description) at position x with the parameters given in params and scalefactors.
	 * 
	 * @param x The position where g_s is evaluated.
	 * @param params The parameters {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3} for the three Gaussians.
	 * @param scalefactors The scaling factors {\alpha_1, \alpha_2, \alpha_3} for the three Gaussians.
	 * @return Returns the function value at position x.
	 */
	static double evaluateTripleGaussian(double x, double[] params, double[] scalefactors) {
		double result = 0;
		for(int i = 0; i < 3; ++i) {
			result += scalefactors[i]*evaluateGaussian(x, params[i*2], params[i*2+1]);
		}
		return result;
	}

	/**
	 * Evaluates an (unscaled) Gaussian function at position x.
	 * @param x
	 * @param mean The mean of the Gaussian function.
	 * @param standardDev The standard deviation of the Gaussian function.
	 * @return Returns the function value at position x.
	 */
	static double evaluateGaussian (double x, double mean, double standardDev) {
		return 1/(standardDev*Math.sqrt(2*Math.PI))*Math.pow(Math.E, -0.5*((x-mean)*(x-mean)/(standardDev*standardDev)));
	}

	/**
	 * Evaluates an scaled Gaussian function at position x.
	 * @param x
	 * @param mean The mean of the Gaussian function.
	 * @param standardDev The standard deviation of the Gaussian function.
	 * @return Returns the function value at position x.
	 */
	static double evaluateGaussian (double x, double mean, double standardDev, double scaling) {
		return scaling * (1/(standardDev*Math.sqrt(2*Math.PI))*Math.pow(Math.E, -0.5*((x-mean)*(x-mean)/(standardDev*standardDev))));
	}

	
	/**
	 * Computes the partial derivative with respect to <b>mean_respGradientNo</b> of the optimization function h_s.
	 * 
	 * @param params The parameters {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3} for the three Gaussians.
	 * @param scalefactors The scaling factors {\alpha_1, \alpha_2, \alpha_3} for the three Gaussians.
	 * @param respGradientNo Denotes which variable is used for the partial derivation (= denotes the index of mean_i \in {0, 1, 2}).
	 * @return The partial derivative of the optimization function given parameters params and scalefactors.
	 */
	protected double calculateGradientRespToMean(double[] params, double scalefactors[], int respGradientNo) {
		double respMean = params[respGradientNo*2];
		double respStandDev = params[respGradientNo*2+1];
		double respAlpha = scalefactors[respGradientNo];
		double gradientResult = 0;
		for(int i = 0; i < yvalues.length; ++i) {
			gradientResult += 2*(yvalues[i]-evaluateTripleGaussian(xvalues[i], params, scalefactors))*-respAlpha*evaluateGaussian(xvalues[i], respMean, respStandDev)*(xvalues[i]-respMean)/(respStandDev*respStandDev);
		}
		return gradientResult;
	}
	
	/**
	 * Computes the partial derivative with respect to <b>standardDev_respGradientNo</b> of the optimization function h_s.
	 * 
	 * @param params The parameters {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3} for the three Gaussians.
	 * @param scalefactors The scaling factors {\alpha_1, \alpha_2, \alpha_3} for the three Gaussians.
	 * @param respGradientNo Denotes which variable is used for the partial derivation (= denotes the index of standardDev_i \in {0, 1, 2}).
	 * @return The partial derivative of the optimization function given parameters params and scalefactors.
	 */
	protected double calculateGradientRespToStandardDev(double[] params, double[] scalefactors, int respGradientNo) {
		double respMean = params[respGradientNo*2];
		double respStandDev = params[respGradientNo*2+1];
		double respAlpha = scalefactors[respGradientNo];
		double gradientResult = 0;
		for(int i = 0; i < yvalues.length; ++i) {
			gradientResult += 2*(yvalues[i]-evaluateTripleGaussian(xvalues[i], params, scalefactors))
					* respAlpha*evaluateGaussian(xvalues[i], respMean, respStandDev) 
							*((1/respStandDev)-((xvalues[i]-respMean)*(xvalues[i]-respMean))/(respStandDev*respStandDev*respStandDev));
		}
		return gradientResult;
	}
	
	/**
	 * Computes the partial derivative with respect to <b>scaleFactor_respGradientNo</b> of the optimization function h_s.
	 * 
	 * @param params The parameters {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3} for the three Gaussians.
	 * @param scalefactors The scaling factors {\alpha_1, \alpha_2, \alpha_3} for the three Gaussians.
	 * @param respGradientNo Denotes which variable is used for the partial derivation (= denotes the index of scaleFactor_i \in {0, 1, 2}).
	 * @return The partial derivative of the optimization function given parameters params and scalefactors.
	 */
	protected double calculateGradientRespToAlpha(double[] params, double[] scalefactors, int respGradientNo) {
		double respMean = params[respGradientNo*2];
		double respStandDev = params[respGradientNo*2+1];
		double gradientResult = 0;
		for(int i = 0; i < yvalues.length; ++i) {
			gradientResult += 2*(yvalues[i]-evaluateTripleGaussian(xvalues[i], params, scalefactors))
							*(-evaluateGaussian(xvalues[i], respMean, respStandDev));
		}
		return gradientResult;
	}

	/**
	 * Evaluates the function g_s (see class description) at position x.
	 * @param x
	 * @param params The parameters for the three Gaussians. params = {\mu_1, \sigma_1, \mu_2, \sigma_2, \mu_3, \sigma_3, \alpha_1, \alpha_2, \alpha_3}
	 */
	public double evaluateBasisFunction(double x, double[] params) {
		if(params.length != 9) {
			throw new IllegalArgumentException("Wrong number of parameters for the evaluation of the function!");
		}
		double[] params_1 = Arrays.copyOf(params, 6);
		double[] scalefactors = {params[6], params[7], params[8]};
		
		return evaluateTripleGaussian(x, params_1, scalefactors);
	}
	
}
