/*
 *  Copyright (C) 2013-2014 Katharina Breininger
 *
 *   This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.gui.Roi;
import ij.measure.Calibration;
import ij.measure.ResultsTable;
import ij.plugin.PlugIn;
import ij.plugin.filter.PlugInFilter;
import ij.plugin.frame.RoiManager;
import ij.process.ImageProcessor;

import java.awt.Polygon;

/**
 * This ImageJ plugIn has been programmed to compare the ROIs constructed by "Vermessung_Nasenfluegel" with a ground truth.
 * For this purpose three pairs of ROIs (left, middle and right wall) are compared and the mean and variance of the differences are computed.
 * Each pair is made up by a ROI computed by "Vermessung_Nasenfluegel" and the corresponding ground truth.
 * As input for this plugIn the IDs of the ROIs w.r.t. the ROIManager are required.
 * If the mean and variance shall be computed w.r.t. pixel distances, then this plugIn does not need to be called with an image.
 * If the mean and variance shall be computed w.r.t. millimeter distances, then this plugIn has to be called with a image corresponding to the ROIs.
 * Right now, only pixels and millimeters are supported as units.
 * The moments are calculated w.r.t. the signed differences as well as the absolute differences.
 * 
 * This is essentially an ImageJ Plugin - to use it change the name so that it includes a underline '_'.
 */
public class ValidateMeasurementsGroundTruth implements PlugInFilter, PlugIn {
	
	/**
	 * Enum to map the possible units pixels and millimeters.
	 */
	enum unit {PIXELS("pixels", "px"), MILLIMETERS("millimeters", "mm");
		final String stringValue;
		final String shortStringValue;
		private unit(final String str, final String shortStr) {
			stringValue = str;
			shortStringValue = shortStr;
		}
		static String[] getArray() {
			unit[] units = unit.values();
			String[] result = new String[units.length];
			for(int i = 0; i < units.length; i++) {
				result[i] = units[i].stringValue;
			}
			return result;
		}
		public String toString() {
			return stringValue;
		}
	}
	
	
	Roi[][] roiPairs;
	int[][][] xvals;
	int[] signeddifferences;
	int[] absdifferences;
	unit unitType;
	ImagePlus img;
	
	private boolean debug = false;

	/* ** SECTION: IMAGEJ METHODS *** */
	@Override
	public void run(String arg) {
		img = null;
		this.run((ImageProcessor)null);
	}

	@Override
	public int setup(String arg, ImagePlus imp) {
		this.img = imp;
		return DOES_ALL + NO_CHANGES;
	}

	@Override
	public void run(ImageProcessor ip) {
		roiPairs = new Roi[3][2];
		if(showDialog()) {
			getValues();
			getDifferences();
			double absMean = calculateMean(absdifferences);
			double signedMean = calculateMean(signeddifferences);
			showResultTable();
			if(debug) {
				System.out.println("Abs. Mean: " + absMean);
				System.out.println("Signed Mean: " + signedMean);
				System.out.println("Abs. Variance: " + calculateVariance(absdifferences, absMean));
				System.out.println("Signed Variance: " + calculateVariance(signeddifferences, signedMean));
			}
		}
		
	}
	
	/**
	 * Opens the dialogue for this plugIn.
	 * You have to put in three pairs of ROIs with same length and type Roi.POLYLINE to use this plugIn.
	 * You can also choose, if you want to have the measurements in pixel or in millimeters. 
	 * If millimeters is chosen, you have to call the plugIn with the corresponding image.
	 * 
	 * @return Returns false, if the dialog has been canceled, true otherwise.
	 */
	public boolean showDialog() {
		GenericDialog gd = new GenericDialog("Plugin options");
		gd.addMessage("Put in field corresponding ROI pairs, separate elements with \",\"");
		gd.addStringField("ROI pair 1", "");
		gd.addStringField("ROI pair 2", "");
		gd.addStringField("ROI pair 3", "");
		gd.addChoice("Unit", unit.getArray(), unit.getArray()[0]);
		gd.showDialog();
		if(gd.wasCanceled())
			return false;
		String[] pairsString = {gd.getNextString(), gd.getNextString(), gd.getNextString()};
		unitType = unit.values()[gd.getNextChoiceIndex()];
		
		if(img == null && unitType != unit.PIXELS) {
			IJ.showMessage("Unit type " + unitType.stringValue + " not supported if no image is available. Unit type is set to " + unit.PIXELS.stringValue +".");
			unitType = unit.PIXELS;
		}
		
		int[][] roiPairIDs = new int[3][2];
		for(int i = 0; i < 3; i++) {
			pairsString[i]= pairsString[i].replaceAll("\\s", "");
			String[] ids = pairsString[i].split(",");
			if(debug) {
				for(String id: ids) {
					System.out.println(id);
				}
			}
			if(ids.length != 2) {
				IJ.error("Invalid number of ROIs in input field - plugin aborted");
				return false;
			}
			try {
				roiPairIDs[i][0] = Integer.parseInt(ids[0]);
				roiPairIDs[i][1] = Integer.parseInt(ids[1]);
			}
			catch(NumberFormatException e) {
				IJ.error("Not able to parse "+ids[0] + " or " + ids[1]);
				return false;
			}
		}
		RoiManager roiManager = RoiManager.getInstance() == null ? new RoiManager() : RoiManager.getInstance() ;
		Roi[] currentRois = roiManager.getRoisAsArray();
		for(int i = 0; i < 3; i++) {
			if(currentRois[roiPairIDs[i][0]].getType() != Roi.POLYLINE || currentRois[roiPairIDs[i][1]].getType() != Roi.POLYLINE) {
				IJ.error("ROI type must be POLYLINE! - plugin aborted");
				return false;
			}
			roiPairs[i][0] = currentRois[roiPairIDs[i][0]];
			roiPairs[i][1] = currentRois[roiPairIDs[i][1]];
		}
		return true;
	}

	/* *** SECTION: IMAGEJ METHODS END *** */

	/**
	 * Translates the given polyline ROIs in integer values. The x-values of each polyline point are put into the array xvals.
	 * xvals has then the following form:
	 * The first ROI in the ith pair is
	 * @return
	 * @throws IllegalArgumentException
	 */
	boolean getValues() throws IllegalArgumentException {
		
		xvals = new int[3][2][];
		
		for(int i = 0; i < 3; i++) {
			Polygon poly0 = roiPairs[i][0].getPolygon();
			Polygon poly1 = roiPairs[i][1].getPolygon();
			xvals[i][0] = new int[poly0.npoints];
			xvals[i][1] = new int[poly1.npoints];
			for(int j = 0; j < poly0.npoints; ++j) {
				xvals[i][0][j] = poly0.xpoints[j];
				xvals[i][1][j] = poly1.xpoints[j];
			}
		}
		return true;

	}

	/**
	 * Calculates the absolute and the signed differences between all values given in xvals[i][0] and xvals[i][1] for i = 0, 1, 2.
	 * Please note that those array x[i][k] must have the same length for all i, k.
	 * The differences are put in absdifferences resp. signeddifferences in the following order:
	 * {
	 * diff(xvals[0][0][0] - xvals[0][1][0]), ..., diff(xvals[0][0][npoints-1] - xvals[0][1][npoints-1]), 
	 * diff(xvals[0][0][0] - xvals[1][1][0]), ..., diff(xvals[1][0][npoints-1] - xvals[1][1][npoints-1]), 
	 * diff(xvals[0][0][0] - xvals[2][1][0]), ..., diff(xvals[2][0][npoints-1] - xvals[2][1][npoints-1])
	 * } 
	 */
	public void getDifferences() {
		if(xvals[0][0].length != xvals[1][0].length || xvals[1][0].length != xvals[2][0].length) {
			IJ.error("Plugin can only handle thre rois with same length. Pluin is aborted.");
		}
		for(int i = 0; i < 3; i++) {
			Polygon poly0 = roiPairs[i][0].getPolygon();
			Polygon poly1 = roiPairs[i][1].getPolygon();
			if(poly0.npoints != poly1.npoints) {
				throw new IllegalArgumentException("Number of points for ROIs does not coincide." + " Polygon 0: " + poly0.npoints + ", Polygon 1: " + poly1.npoints +".\n"
						+ " Problem for ids " + roiPairs[i][0] + " and " + roiPairs[i][1] + ".");				
			}
		}
		int npoints = xvals[0][0].length;
		absdifferences = new int[3*npoints];
		signeddifferences = new int[3*npoints];
		for(int i = 0; i < 3; i++) {
			for(int j = 0; j < npoints; ++j) {
				int distance = xvals[i][0][j] - xvals[i][1][j]; // distance in pixels
				if(unitType == unit.MILLIMETERS) {	
					Calibration calibr = img.getCalibration();
					if(calibr.getXUnit().equals("mm"))
						distance *= calibr.pixelWidth; // calculate distance in mm if necessary
				}
				absdifferences[j+i*npoints] = Math.abs(distance);
				signeddifferences[j+i*npoints] = distance;
			}
		}
	}
	
	/**
	 * Calculates the mean of the given (integer) values. Given (x1, ..., xn) the mean \mu is calculated using \mu = 1/n * \sum_{i=1}^n.
	 * @param values An array with Integer values.
	 * @return The mean of all values in the array.
	 */
	public double calculateMean(int[] values) {
		double mean = 0;
		for(int i = 0; i < values.length; i++) {
			mean += values[i];
		}
		return mean/values.length;
	}
	
	/**
	 * Calculates the variance of the given (integer) values. 
	 * @param values An array with integer values.
	 * @return The variance of all values in the array.
	 */
	public double calculateVariance(int[] values) {
		double mean = calculateMean(values);
		return calculateVariance(values, mean);
	}
	
	/**
	 * Calculates the variance of the given (integer) values.
	 * If the mean has already been calculated for the array values, this method can be used.
	 * @param values An array with integer values.
	 * @return The variance of all values in the array.
	 */
	public double calculateVariance(int[] values, double mean) {
		double variance = 0;
		for(int i = 0; i < values.length; ++i) {
		 variance += (values[i] - mean)*(values[i] - mean);
		}
		return variance/values.length;
	}
	
	/**
	 * Method displays the result table and the calculated values for mean and variance.
	 */
	public void showResultTable() {
		ResultsTable resultsTable = new ResultsTable();
		resultsTable.incrementCounter();
		double absMean = calculateMean(absdifferences);
		double absVariance = calculateVariance(absdifferences, absMean);
		double signedMean = calculateMean(signeddifferences);
		double signedVariance = calculateVariance(signeddifferences, signedMean);
		String unitStr = unitType.shortStringValue;
		resultsTable.addValue("Mean (abs. diff.) " + unitStr, absMean);
		resultsTable.addValue("Variance (abs. diff.) " + unitStr, absVariance);
		resultsTable.addValue("Mean (sign. diff.) " + unitStr, signedMean);
		resultsTable.addValue("Variance (sign. diff.) " + unitStr, signedVariance);
		resultsTable.showRowNumbers(false);
		resultsTable.show("Moments");
	}

}